/*
	FrameParser - author: Daniel Roggen
	Version 1.1
*/

#include <stdio.h>
//#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>

#include "FrameParser.h"



int FrameParser(const char *format,const char *frame,...)
{
	unsigned frameheaderlen,channel;
	char type,*t;
	unsigned int *vaptr;
	const unsigned char *dataptr;
	unsigned int dataunsigned;
	const char *formatptr;
	

	va_list ap;
	va_start(ap,frame);

	
	
	// Find frame header length
	t=(char*)strchr(format,';');
	if(t==0)
	{
		assert(t);
		return FRAMEPARSER_ERROR_FORMAT;
	}
	frameheaderlen=t-format;

	//printf("Length of header: %d\n",frameheaderlen);

	// Iterate the channels and decode the data
	dataptr = ((const unsigned char*)frame)+frameheaderlen;
	formatptr=format+frameheaderlen+1;
	channel=0;
	while(1)
	{
		
		bool issigned;
		// Get next data type
		type = *formatptr++;
		if(type==0)
			break;		// Last item processed -> return.
		if(type=='-')
		{
			// Signed element. We have to read the data type.
			issigned=true;
			type = *formatptr++;
			if(type==0)
			{
				va_end(ap);
				assert(0);
				return FRAMEPARSER_ERROR_FORMAT;		// Error: we have a sign, but no type specifier -> crash.
			}
		}
		else
			issigned=false;
		

		// Assume all arguments passed as 32-bits - that's a documentation requirement.
		vaptr = va_arg(ap,unsigned int *);		

		switch(type)
		{
			case 'c':
				dataunsigned = *dataptr++;
				if(issigned && dataunsigned&0x80)
					dataunsigned|=0xFFFFFF00;
				break;
			case 's':
				dataunsigned = *dataptr++;
				dataunsigned |= (*dataptr++)<<8;
				if(issigned && dataunsigned&0x8000)
					dataunsigned|=0xFFFF0000;
				break;
			case 'S':
				dataunsigned = (*dataptr++)<<8;
				dataunsigned |= *dataptr++;
				if(issigned && dataunsigned&0x8000)
					dataunsigned|=0xFFFF0000;
				break;
			case 'i':
				dataunsigned = *dataptr++;
				dataunsigned |= (*dataptr++)<<8;
				dataunsigned |= (*dataptr++)<<16;
				dataunsigned |= (*dataptr++)<<24;
				break;
			case 'I':
				dataunsigned = (*dataptr++)<<24;
				dataunsigned |= (*dataptr++)<<16;
				dataunsigned |= (*dataptr++)<<8;
				dataunsigned |= *dataptr++;
				break;
			default:
				va_end(ap);
				assert(0);
				return FRAMEPARSER_ERROR_FORMAT;
		}
		*vaptr=dataunsigned;
		channel++;
	}	// do
	
	va_end(ap);
	return FRAMEPARSER_NOERROR;
}
int FrameParserV(const char *format,const char *frame,unsigned int *vector)
{
	unsigned frameheaderlen,channel;
	char type,*t;
	const unsigned char *dataptr;
	unsigned int dataunsigned;
	const char *formatptr;
	unsigned int *vaptr;
	
	
	// Find frame header length
	t=(char*)strchr(format,';');
	if(t==0)
	{
		assert(t);
		return FRAMEPARSER_ERROR_FORMAT;
	}
	frameheaderlen=t-format;

	//printf("Length of header: %d\n",frameheaderlen);

	// Iterate the channels and decode the data
	dataptr = ((const unsigned char*)frame)+frameheaderlen;
	formatptr=format+frameheaderlen+1;
	channel=0;
	while(1)
	{
		
		bool issigned;
		// Get next data type
		type = *formatptr++;
		if(type==0)
			break;		// Last item processed -> return.
		if(type=='-')
		{
			// Signed element. We have to read the data type.
			issigned=true;
			type = *formatptr++;
			if(type==0)
			{
				assert(0);
				return FRAMEPARSER_ERROR_FORMAT;		// Error: we have a sign, but no type specifier -> crash.
			}
		}
		else
			issigned=false;
		
		switch(type)
		{
			case 'c':
				dataunsigned = *dataptr++;
				if(issigned && dataunsigned&0x80)
					dataunsigned|=0xFFFFFF00;
				break;
			case 's':
				dataunsigned = *dataptr++;
				dataunsigned |= (*dataptr++)<<8;
				if(issigned && dataunsigned&0x8000)
					dataunsigned|=0xFFFF0000;
				break;
			case 'S':
				dataunsigned = (*dataptr++)<<8;
				dataunsigned |= *dataptr++;
				if(issigned && dataunsigned&0x8000)
					dataunsigned|=0xFFFF0000;
				break;
			case 'i':
				dataunsigned = *dataptr++;
				dataunsigned |= (*dataptr++)<<8;
				dataunsigned |= (*dataptr++)<<16;
				dataunsigned |= (*dataptr++)<<24;
				break;
			case 'I':
				dataunsigned = (*dataptr++)<<24;
				dataunsigned |= (*dataptr++)<<16;
				dataunsigned |= (*dataptr++)<<8;
				dataunsigned |= *dataptr++;
				break;
			default:
				assert(0);
				return FRAMEPARSER_ERROR_FORMAT;
		}
		vector[channel]=dataunsigned;
		channel++;
	}	// do
	
	return FRAMEPARSER_NOERROR;
}
int FrameParser_GetFrameSize(const char *format)
{
	int c,s;
	int rv;
	rv=FrameParser_Stats(format,c,s);
	if(rv!=FRAMEPARSER_NOERROR)
		return rv;
	return s;
}
int FrameParser_GetNumChannels(const char *format)
{
	int c,s;
	int rv;
	rv=FrameParser_Stats(format,c,s);
	if(rv!=FRAMEPARSER_NOERROR)
		return rv;
	return c;
}

int FrameParser_ValidateFormat(const char *format)
{
	unsigned frameheaderlen;
	char type,*t;
	const char *formatptr;

// Find frame header length
	t=(char*)strchr(format,';');
	if(t==0)
	{
		return FRAMEPARSER_ERROR_FORMAT;
	}
	frameheaderlen=t-format;

	// Decode the format string
	formatptr=format+frameheaderlen+1;
	while(1)
	{
		// Get next data type
		type = *formatptr++;
		if(type==0)
			break;		// Last item processed -> return.
		if(type=='-')
		{
			type = *formatptr++;
			if(type==0)
			{
				assert(0);
				return FRAMEPARSER_ERROR_FORMAT;
			}
		}
		if(type!='c' && type !='s' && type !='S' && type !='i' && type !='I')
			return FRAMEPARSER_ERROR_FORMAT;
	}	// do
	return FRAMEPARSER_NOERROR;
}
int FrameParser_IsFrame(const char *format,const char *stream)
{
	unsigned frameheaderlen;
	char type,*t;
	const char *formatptr;

// Find frame header length
	t=(char*)strchr(format,';');
	if(t==0)
	{
		assert(t);
		return FRAMEPARSER_ERROR_FORMAT;
	}
	frameheaderlen=t-format;
	
	if(strncmp(stream,format,frameheaderlen) == 0)
		return 1;
	return 0;
}
int FrameParser_IsSigned(const char *format,int c)
{
	unsigned int channel;
	unsigned frameheaderlen;
	char type,*t;
	const char *formatptr;

// Find frame header length
	t=(char*)strchr(format,';');
	if(t==0)
	{
		return FRAMEPARSER_ERROR_FORMAT;
	}
	frameheaderlen=t-format;

	// Decode the format string
	formatptr=format+frameheaderlen+1;
	channel=0;
	while(1)
	{
		bool issigned;
		// Get next data type
		type = *formatptr++;
		if(type==0)
			break;		// Last item processed -> return.
		if(type=='-')
		{
			issigned=true;
			// Signed element. We have to read the data type.
			type = *formatptr++;
			if(type==0)
			{
				assert(0);
				return FRAMEPARSER_ERROR_FORMAT;		// Error: we have a sign, but no type specifier -> crash.
			}
		}
		else
			issigned=false;
		// Don't care about type; just return signed or not.
		if(channel==c)
			return issigned;
		
		channel++;
	}	// do
	return FRAMEPARSER_ERROR_FORMAT;
}
int FrameParser_Stats(const char *format,int &channel,int &framesize)
{
	unsigned frameheaderlen;
	char type,*t;
	const char *formatptr;

// Find frame header length
	t=(char*)strchr(format,';');
	if(t==0)
	{
		return FRAMEPARSER_ERROR_FORMAT;
	}
	frameheaderlen=t-format;

	// Decode the format string
	formatptr=format+frameheaderlen+1;
	channel=0;
	framesize=frameheaderlen;
	while(1)
	{
		// Get next data type
		type = *formatptr++;
		if(type==0)
			break;		// Last item processed -> return.
		if(type=='-')
		{
			// Signed element. We have to read the data type.
			type = *formatptr++;
			if(type==0)
			{
				assert(0);
				return FRAMEPARSER_ERROR_FORMAT;		// Error: we have a sign, but no type specifier -> crash.
			}
		}
	
		switch(type)
		{
			case 'c':
				framesize+=1;
				break;
			case 's':
			case 'S':
				framesize+=2;
				break;
			case 'i':
			case 'I':
				framesize+=4;
				break;
			default:
				assert(0);
				return FRAMEPARSER_ERROR_FORMAT;
		}
		channel++;
	}	// do
	return FRAMEPARSER_NOERROR;
}









